#!/usr/bin/env python
"""Report memory usage during chunking of a large file
"""
# from pyaddons import utils
from merge_sort import chunks
import gzip
import csv
import os
import resource
import tempfile

infile = "/data/mapping_files/b38/v20220402/gwas_norm.v20220402.common.biallelic.vep.b38.vcf.gz"
delimiter = "\t"

with gzip.open(infile, 'rt') as infile:
    line = '##'
    while line.startswith('##'):
        line = infile.readline()

    with open(os.path.join(os.environ['HOME'], "merge_mem_usage.txt"), 'wt') as memhf:
        idx = 1
        reader = csv.reader(infile, delimiter=delimiter)
        chunk_dir = tempfile.mkdtemp(prefix='merge_sort_test_')
        chunksize = 100000
        with chunks.CsvSortedChunks(
                chunk_dir, lambda x: (x[0], int(x[1]), x[3], x[4]),
                chunksize=chunksize, delimiter=delimiter
        ) as chunker:
            for row in reader:
                if (idx % 100000) == 0:
                    mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
                    print(f"{idx}: main max mem usage: {round(mem/1024, 2)} MB")
                idx += 1
                chunker.add_row(row)



"""Classes for writing content to sorted intermediate files

There are a several classes providing the same interface just different
implementations so I can understand what technique is best.
"""
from merge_sort import common
from pyaddons import utils
from sortedcontainers import SortedList
import gzip
import os
import csv
import pprint as pp


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _BaseChunks(object):
    """Base class for Write files that are sorted on key where each file has
    at max chunksize rows.

    Parameters
    ----------
    chunk_dir : `str`
        The directory to write the chunk files to.
    key : `function`
        A key function that will return the values to sort on. This will be
        passed onto the sorting routine.
    chunksize : `int`, optional, default: `100000`
        The number of rows to store in memory before sorting and outputting to
        the chunk file.
    write_mode : `function`, optional, default: `wt`
        The method mode to use to write the chunk files. In most cases this
        should not be changed but it is there just in case the ``write_method``
        has any odd requirements.
    write_method : `function`, optional, default: `gzip.open`
        The method to use to write the chunk files.
    header : `list` of `str`, optional, default: `NoneType`
        An optional header to write to the start of each chunk file.
    delete_on_error : `bool`, optional, default: `True`
        Delete all the chunk files if the chunking routine ends in an error.
    reverse : `bool`, optional, default: `False`
        Sort in reverse.
    chunk_prefix : `str`, optional, default: `NoneType`
        An optional prefix to add to the start of each chunk file name.
    chunk_suffix : `str`, optional, default: `.chunk.gz`
        An optional suffix to add to to the end of each chunk filename. Set to
        ``NoneType`` if you do not want a suffix.
    **write_kwargs
        Any keyword arguments that will be passed onto the ``write_method``.
    """

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, chunk_dir, key=None, chunksize=100000, write_mode='wt',
                 write_method=gzip.open, header=None, delete_on_error=True,
                 reverse=False, chunk_prefix=None, chunk_suffix=".chunk.gz",
                 **write_kwargs):
        self._chunk_dir = chunk_dir
        self._key = key
        self._chunksize = None
        self._write_mode = write_mode
        self._write_method = write_method
        self._header = header
        self._delete_on_error = delete_on_error
        self._reverse = None
        self._chunk_prefix = chunk_prefix
        self._chunk_suffix = chunk_suffix
        self._write_kwargs = write_kwargs
        self.reverse = reverse
        self.chunksize = chunksize

        self._chunks = {}
        self._tmp_file = None
        self._cur_chunk = self.get_empty_chunk()
        self._idx = 0
        self._n_added = 0
        self._n_written = 0

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __enter__(self):
        """Entry point for the context manager.
        """
        return self.open()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __exit__(self, *args):
        """Exit point for the context manager

        Parameters
        ----------
        *args
            A tuple containing any errors that may have caused the context
            manager to exit early. If no errors were generated and the context
            manager exited normally the tuple will be (None, None, None).
        """
        self.close()

        if args[0] is not None and self._delete_on_error is True:
            # TODO: Delete all the chunk files if we error
            self.delete()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def open(self):
        """Open the input file and setup everything for reading.

        Returns
        -------
        self : `merge_sort.chunks._BaseChunks`
            For chaining.
        """
        self.open_chunk_file()
        self.write_header()
        return self

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def close(self):
        """Close all files.
        """
        self.flush_chunk()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def delete(self):
        """Delete all the chunk files created by the object.

        If the chunk file does not exist, this will fail silently, other errors
        will be raised though.
        """
        for i in self.chunk_file_list:
            try:
                os.unlink(i)
            except FileNotFoundError:
                # Might have already been deleted by something else, if used in
                # an external merge sort
                pass

        self._chunks = {}

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def add_row(self, row):
        """Add a row to the current chunk.

        Parameters
        ----------
        row : `str`
            The row to add to the file. Should contain a newline at the end.
        """
        self._n_added += 1

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def chunk_files(self):
        """Get all the chunk file names (`dict`).
        """
        return self._chunks

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def chunk_file_list(self):
        """Get all the chunk file names in the order they were written
        (`list` of `str`).
        """
        return list(self._chunks.keys())

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def n_chunks(self):
        """Get the number of chunks that have been written to file (this does
        not include the current chunk) (`int`)
        """
        return len(self._chunks)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def n_added(self):
        """Get the number of rows in the current chunk (`int`).
        """
        return self._n_added

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def n_written(self):
        """Get the number of rows in the current chunk (`int`).
        """
        return self._n_written

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def n_written_total(self):
        """Get the total number of rows written to all chunks
        (not including any current chunks) (`int`).
        """
        return sum(self._chunks.values())

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def current_chunk_file(self):
        """Get the current chunk file object (`File`).
        """
        return self._tmp_file

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def current_chunk(self):
        """Get the current chunk (`list`).
        """
        return self._cur_chunk

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def reverse(self):
        """Get the reverse argument (`bool`).
        """
        return self._reverse

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def header(self):
        """Get the header argument (`bool`).
        """
        return self._header

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reverse.setter
    def reverse(self, reverse):
        """Set the reverse argument (`bool`).
        """
        if not isinstance(reverse, bool):
            raise TypeError(
                f"reverse should be a Boolean, not: {type(reverse)}"
            )
        self._reverse = reverse

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def chunksize(self):
        """Get the chunksize argument, should be a positive integer (`int`).
        """
        return self._chunksize

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @chunksize.setter
    def chunksize(self, chunksize):
        """Set the chunksize argument, should be a positive integer (`int`).
        """
        if not isinstance(chunksize, int):
            raise TypeError(
                f"chunksize should be a positive integer, not: {chunksize}"
            )

        if chunksize <= 0:
            raise ValueError(
                f"chunksize should be a positive integer, not: {chunksize}"
            )
        self._chunksize = chunksize

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def key(self):
        """Get the sort key argument (`function`).
        """
        return self._key

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @chunksize.setter
    def chunksize(self, chunksize):
        """Set the chunksize argument, should be a positive integer (`int`).
        """
        if not isinstance(chunksize, int):
            raise TypeError(
                f"chunksize should be a positive integer, not: {chunksize}"
            )

        if chunksize <= 0:
            raise ValueError(
                f"chunksize should be a positive integer, not: {chunksize}"
            )
        self._chunksize = chunksize

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def set_new_chunk_file(self):
        """Initialise a new chunk file to write the chunk to.

        Raises
        ------
        FileExistsError
            If any existing chunk files have not been closed first.
        """
        if self._tmp_file is not None:
            raise FileExistsError("please close existing chunk file first")

        # Create a new one
        self._tmp_file = utils.get_tmp_file(
            dir=self._chunk_dir, prefix=self._chunk_prefix,
            suffix=self._chunk_suffix
        )

        self._tmp_file = self._write_method(
            self._tmp_file, self._write_mode, **self._write_kwargs
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def open_chunk_file(self):
        """Open a chunk file.

        Returns
        -------
        chunk_file: `File`
            The file object for the chunk.

        Raises
        ------
        FileExistsError
            If any existing chunk files have not been closed first.
        """
        self.set_new_chunk_file()
        return self._tmp_file

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def close_chunk_file(self):
        """Close a tmp chunk file

        If one is not initialise then it will fail silently. The closed chunk
        file name will be added to the _chunks dict only if it has been written
        to.
        """
        try:
            self._tmp_file.close()
        except AttributeError:
            return

        if self._n_written > 0:
            self._chunks[self._tmp_file.name] = self._n_written
        else:
            # remove the empty chunk file
            os.unlink(self._tmp_file.name)

        # Empty the file
        self._tmp_file = None
        self._n_written = 0

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def write_header(self):
        """write the header row to the chunk file.
        """
        if self._header is not None:
            self.write_row(self.header)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def sort_chunk(self):
        """Sort the chunk.
        """
        pass

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_empty_chunk(self):
        """Get an empty chunk holder.

        Returns
        -------
        empty_chunk : `list`
            An empty chunk to fill.
        """
        return []

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def clear_chunk(self):
        """Clear the contents of the chunk.
        """
        self._cur_chunk = self.get_empty_chunk()
        self._n_added = 0

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def write_chunk(self):
        """Write the chunk to file.
        """
        for row in self._cur_chunk:
            self.write_row(row)
            self._n_written += 1

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def write_row(self, row):
        """Write a single row to the chunk file.

        Parameters
        ----------
        row : `str`
            The row to write to the file. The row should contain a newline
            character.
        """
        self._tmp_file.write((str(row)))

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def flush_chunk(self):
        """Perform the whole process of sorting,opening,wring,closing and
        clearing a chunk.
        """
        pass


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class SortedChunks(_BaseChunks):
    """Write files that are sorted on key where each file has at max
    chunksize rows.

    Parameters
    ----------
    chunk_dir : `str`
        The directory to write the chunk files to.
    key : `function`
        A key function that will return the values to sort on. This will be
        passed onto the sorting routine.
    chunksize : `int`, optional, default: `100000`
        The number of rows to store in memory before sorting and outputting to
        the chunk file.
    write_mode : `function`, optional, default: `wt`
        The method mode to use to write the chunk files. In most cases this
        should not be changed but it is there just in case the ``write_method``
        has any odd requirements.
    write_method : `function`, optional, default: `gzip.open`
        The method to use to write the chunk files.
    header : `str`, optional, default: `NoneType`
        An optional header to write to the start of each chunk file.
    delete_on_error : `bool`, optional, default: `True`
        Delete all the chunk files if the chunking routine ends in an error.
    reverse : `bool`, optional, default: `False`
        Sort in reverse.
    chunk_prefix : `str`, optional, default: `NoneType`
        An optional prefix to add to the start of each chunk file name.
    chunk_suffix : `str`, optional, default: `.chunk.gz`
        An optional suffix to add to to the end of each chunk filename. Set to
        ``NoneType`` if you do not want a suffix.
    **write_kwargs
        Any keyword arguments that will be passed onto the ``write_method``.
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def sort_chunk(self):
        """Sort the chunk.
        """
        self.current_chunk.sort(key=self.key, reverse=self.reverse)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def add_row(self, row):
        """Add a row to the current chunk.

        Parameters
        ----------
        row : `str`
            The row to add to the file. Should contain a newline at the end.
        """
        super().add_row(row)
        self.current_chunk.append(row)

        if self.n_added == self.chunksize:
            self.flush_chunk()
            self.open_chunk_file()
            self.write_header()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def flush_chunk(self):
        """Perform the whole process of sorting,opening,wring,closing and
        clearing a chunk.
        """
        self.sort_chunk()
        self.write_chunk()
        self.close_chunk_file()
        self.clear_chunk()


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class SortedListChunks(_BaseChunks):
    """Write files that are sorted on key where each file has at max
    chunksize rows.

    This implementation uses sortedcontainers rather than lists.

    Parameters
    ----------
    chunk_dir : `str`
        The directory to write the chunk files to.
    key : `function`
        A key function that will return the values to sort on. This will be
        passed onto the sorting routine.
    chunksize : `int`, optional, default: `100000`
        The number of rows to store in memory before sorting and outputting to
        the chunk file.
    write_mode : `function`, optional, default: `wt`
        The method mode to use to write the chunk files. In most cases this
        should not be changed but it is there just in case the ``write_method``
        has any odd requirements.
    write_method : `function`, optional, default: `gzip.open`
        The method to use to write the chunk files.
    header : `str`, optional, default: `NoneType`
        An optional header to write to the start of each chunk file.
    delete_on_error : `bool`, optional, default: `True`
        Delete all the chunk files if the chunking routine ends in an error.
    reverse : `bool`, optional, default: `False`
        Sort in reverse.
    chunk_prefix : `str`, optional, default: `NoneType`
        An optional prefix to add to the start of each chunk file name.
    chunk_suffix : `str`, optional, default: `.chunk.gz`
        An optional suffix to add to to the end of each chunk filename. Set to
        ``NoneType`` if you do not want a suffix.
    **write_kwargs
        Any keyword arguments that will be passed onto the ``write_method``.
    """

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_empty_chunk(self):
        """Get an empty chunk holder.

        Returns
        -------
        empty_chunk : `list`
            An empty chunk to fill.
        """
        return SortedList([], key=self.key)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def add_row(self, row):
        """Add a row to the current chunk.

        Parameters
        ----------
        row : `str`
            The row to add to the file. Should contain a newline at the end.
        """
        super().add_row(row)
        self.current_chunk.add(row)

        if self.n_added == self.chunksize:
            self.flush_chunk()
            self.open_chunk_file()
            self.write_header()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def flush_chunk(self):
        """Perform the whole process of sorting,opening,wring,closing and
        clearing a chunk.
        """
        self.sort_chunk()
        self.write_chunk()
        self.close_chunk_file()
        self.clear_chunk()


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class ExtendedChunks(_BaseChunks):
    """Write files that are sorted on key where each file can have variable
    number of rows possibly exceeding the chunksize value.

    Parameters
    ----------
    chunk_dir : `str`
        The directory to write the chunk files to.
    key : `function`
        A key function that will return the values to sort on. This will be
        passed onto the sorting routine.
    chunksize : `int`, optional, default: `100000`
        The number of rows to store in memory before sorting and outputting to
        the chunk file.
    write_mode : `function`, optional, default: `wt`
        The method mode to use to write the chunk files. In most cases this
        should not be changed but it is there just in case the ``write_method``
        has any odd requirements.
    write_method : `function`, optional, default: `gzip.open`
        The method to use to write the chunk files.
    header : `str`, optional, default: `NoneType`
        An optional header to write to the start of each chunk file.
    delete_on_error : `bool`, optional, default: `True`
        Delete all the chunk files if the chunking routine ends in an error.
    reverse : `bool`, optional, default: `False`
        Sort in reverse.
    chunk_prefix : `str`, optional, default: `NoneType`
        An optional prefix to add to the start of each chunk file name.
    chunk_suffix : `str`, optional, default: `.chunk.gz`
        An optional suffix to add to to the end of each chunk filename. Set to
        ``NoneType`` if you do not want a suffix.
    **write_kwargs
        Any keyword arguments that will be passed onto the ``write_method``.

    Notes
    -----
    This will keep adding chunks to a chunk file if the sort order is
    maintained between the last element of the previous chunk and the first
    element of the current chunk. This can reduce the number of chunk files
    that are needed and is useful if you suspect that the input may be sorted
    already.
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._last_write = None

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def close(self):
        """Close the chunker and any open chunk files.
        """
        super().close()
        self.close_chunk_file()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def sort_chunk(self):
        """Sort the chunk.
        """
        self._cur_chunk.sort(key=self.key, reverse=self.reverse)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def add_row(self, row):
        """Add a row to the current chunk.

        Parameters
        ----------
        row : `str`
            The row to add to the file. Should contain a newline at the end.
        """
        super().add_row(row)
        self.current_chunk.append(row)

        if self.n_added == self.chunksize:
            self.flush_chunk()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def flush_chunk(self):
        """Perform the whole process of sorting,opening,wring,closing and
        clearing a chunk.
        """
        self.sort_chunk()

        try:
            # Put together a list to evaluate if this chunk continues from the
            # previous one
            ref = [self._last_write, self._cur_chunk[0]]
        except IndexError:
            # The index error probably means that the current chunk is empty so
            # we check that and if not we will raise
            if len(self.current_chunk) > 0:
                raise
            # Nothing left to do as there is no data in the current chunk
            return

        try:
            result = sorted(ref, key=self.key, reverse=self.reverse)
        except Exception:
            # The type errpr is raised if the last write is not initialised
            # which will happen on the first write to the first file, but we
            # will check to make sure
            if self._last_write is not None:
                raise
            result = ref

        # Is the sorted list the same as the boundry list
        # TODO: Should I just compare the first element? Better to compare the
        #  actual keys
        if ref == result:
            # extend
            self.write_chunk()
        else:
            # write a new file
            self.close_chunk_file()
            self.open_chunk_file()
            self.write_header()
            self.write_chunk()

        self._last_write = self.current_chunk[-1]
        self.clear_chunk()


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class CsvMixin(object):
    """Add CSV writing abilities to the various chunkers.
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def initialise(self, kwargs):
        """Initialise the mixin with the CSV keyword arguments.
        """
        self.writer = None
        self.csv_kwargs = common.extract_csv_kwargs(kwargs)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def open_chunk_file(self):
        """Open a chunk file.

        Returns
        -------
        chunk_filename : `str`
            The name of the chunk file that has been opened.

        Raises
        ------
        FileExistsError
            If any existing chunk files have not been closed first.
        """
        self.set_new_chunk_file()
        self.writer = csv.writer(self.current_chunk_file, **self.csv_kwargs)
        return self.current_chunk_file

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def write_row(self, row):
        """Write a row to the current chunk.

        Parameters
        ----------
        row : `list`
            The row to add to the file.
        """
        self.writer.writerow(row)


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class CsvSortedChunks(CsvMixin, SortedChunks):
    """Write files that are sorted on key where each file has at max
    chunksize rows. This uses a csv.writer so rows should be added as lists.

    Parameters
    ----------
    chunk_dir : `str`
        The directory to write the chunk files to.
    key : `function`
        A key function that will return the values to sort on. This will be
        passed onto the sorting routine.
    chunksize : `int`, optional, default: `100000`
        The number of rows to store in memory before sorting and outputting to
        the chunk file.
    write_mode : `function`, optional, default: `wt`
        The method mode to use to write the chunk files. In most cases this
        should not be changed but it is there just in case the ``write_method``
        has any odd requirements.
    write_method : `function`, optional, default: `gzip.open`
        The method to use to write the chunk files.
    header : `list` of `str`, optional, default: `NoneType`
        An optional header to write to the start of each chunk file.
    delete_on_error : `bool`, optional, default: `True`
        Delete all the chunk files if the chunking routine ends in an error.
    reverse : `bool`, optional, default: `False`
        Sort in reverse.
    chunk_prefix : `str`, optional, default: `NoneType`
        An optional prefix to add to the start of each chunk file name.
    chunk_suffix : `str`, optional, default: `.chunk.gz`
        An optional suffix to add to to the end of each chunk filename. Set to
        ``NoneType`` if you do not want a suffix.
    **write_kwargs
        Any keyword arguments that will be passed onto the ``write_method``.
    **csv_kwargs
        Any keyword arguments that will be passed onto csv.writer.
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, *args, **kwargs):
        self.initialise(kwargs)
        super().__init__(*args, **kwargs)


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class CsvSortedListChunks(CsvMixin, SortedListChunks):
    """Write files that are sorted on key where each file has at max
    chunksize rows.

    This implementation uses sortedcontainers rather than lists. This uses a
    csv.writer so rows should be added as lists.


    Parameters
    ----------
    chunk_dir : `str`
        The directory to write the chunk files to.
    key : `function`
        A key function that will return the values to sort on. This will be
        passed onto the sorting routine.
    chunksize : `int`, optional, default: `100000`
        The number of rows to store in memory before sorting and outputting to
        the chunk file.
    write_mode : `function`, optional, default: `wt`
        The method mode to use to write the chunk files. In most cases this
        should not be changed but it is there just in case the ``write_method``
        has any odd requirements.
    write_method : `function`, optional, default: `gzip.open`
        The method to use to write the chunk files.
    header : `list` of `str`, optional, default: `NoneType`
        An optional header to write to the start of each chunk file.
    delete_on_error : `bool`, optional, default: `True`
        Delete all the chunk files if the chunking routine ends in an error.
    reverse : `bool`, optional, default: `False`
        Sort in reverse.
    chunk_prefix : `str`, optional, default: `NoneType`
        An optional prefix to add to the start of each chunk file name.
    chunk_suffix : `str`, optional, default: `.chunk.gz`
        An optional suffix to add to to the end of each chunk filename. Set to
        ``NoneType`` if you do not want a suffix.
    **write_kwargs
        Any keyword arguments that will be passed onto the ``write_method``.
    **csv_kwargs
        Any keyword arguments that will be passed onto csv.writer.
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, *args, **kwargs):
        self.initialise(kwargs)
        super().__init__(*args, **kwargs)


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class CsvExtendedChunks(CsvMixin, ExtendedChunks):
    """Write files that are sorted on key where each file can have variable
    number of rows possibly exceeding the chunksize value.

    This uses a csv.writer so rows should be added as lists.

    Parameters
    ----------
    chunk_dir : `str`
        The directory to write the chunk files to.
    key : `function`
        A key function that will return the values to sort on. This will be
        passed onto the sorting routine.
    chunksize : `int`, optional, default: `100000`
        The number of rows to store in memory before sorting and outputting to
        the chunk file.
    write_mode : `function`, optional, default: `wt`
        The method mode to use to write the chunk files. In most cases this
        should not be changed but it is there just in case the ``write_method``
        has any odd requirements.
    write_method : `function`, optional, default: `gzip.open`
        The method to use to write the chunk files.
    header : `list` of `str`, optional, default: `NoneType`
        An optional header to write to the start of each chunk file.
    delete_on_error : `bool`, optional, default: `True`
        Delete all the chunk files if the chunking routine ends in an error.
    reverse : `bool`, optional, default: `False`
        Sort in reverse.
    chunk_prefix : `str`, optional, default: `NoneType`
        An optional prefix to add to the start of each chunk file name.
    chunk_suffix : `str`, optional, default: `.chunk.gz`
        An optional suffix to add to to the end of each chunk filename. Set to
        ``NoneType`` if you do not want a suffix.
    **write_kwargs
        Any keyword arguments that will be passed onto the ``write_method``.
    **csv_kwargs
        Any keyword arguments that will be passed onto csv.writer.

    Notes
    -----
    This will keep adding chunks to a chunk file if the sort order is
    maintained between the last element of the previous chunk and the first
    element of the current chunk. This can reduce the number of chunk files
    that are needed and is useful if you suspect that the input may be sorted
    already.
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, *args, **kwargs):
        self.initialise(kwargs)
        super().__init__(*args, **kwargs)


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class CsvDictMixin(object):
    """Add CSV dict writer abilities to the various chunkers.
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def initialise(self, kwargs):
        """Initialise the mixin with the CSV dict writer keyword arguments.
        """
        self.writer = None
        self.csv_kwargs = common.extract_csv_kwargs(kwargs)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def open_chunk_file(self):
        """Open a chunk file.

        Returns
        -------
        chunk_filename : `str`
            The name of the chunk file that has been opened.

        Raises
        ------
        FileExistsError
            If any existing chunk files have not been closed first.
        """
        if self.header is None:
            raise ValueError(
                "The dictionary mixin requires a header to be available"
            )

        self.set_new_chunk_file()
        self.writer = csv.DictWriter(
            self.current_chunk_file, self.header, **self.csv_kwargs
        )
        return self.current_chunk_file

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def write_header(self):
        """Write a header to the file.
        """
        self.writer.writeheader()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def write_row(self, row):
        """Write a row to the current chunk.

        Parameters
        ----------
        row : `dict`
            The row to add to the file.
        """
        try:
            self.writer.writerow(row)
        except Exception:
            pp.pprint(row)
            raise


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class CsvDictSortedChunks(CsvDictMixin, SortedChunks):
    """Write files that are sorted on key where each file has at max
    chunksize rows.

    This uses a ``csv.DictWriter`` so rows should be added as dictionaries. The
    ``header`` is absolutely required.

    Parameters
    ----------
    chunk_dir : `str`
        The directory to write the chunk files to.
    key : `function`
        A key function that will return the values to sort on. This will be
        passed onto the sorting routine.
    chunksize : `int`, optional, default: `100000`
        The number of rows to store in memory before sorting and outputting to
        the chunk file.
    write_mode : `function`, optional, default: `wt`
        The method mode to use to write the chunk files. In most cases this
        should not be changed but it is there just in case the ``write_method``
        has any odd requirements.
    write_method : `function`, optional, default: `gzip.open`
        The method to use to write the chunk files.
    header : `list` of `str`, optional, default: `NoneType`
        An required header to write to the start of each chunk file.
    delete_on_error : `bool`, optional, default: `True`
        Delete all the chunk files if the chunking routine ends in an error.
    reverse : `bool`, optional, default: `False`
        Sort in reverse.
    chunk_prefix : `str`, optional, default: `NoneType`
        An optional prefix to add to the start of each chunk file name.
    chunk_suffix : `str`, optional, default: `.chunk.gz`
        An optional suffix to add to to the end of each chunk filename. Set to
        ``NoneType`` if you do not want a suffix.
    **write_kwargs
        Any keyword arguments that will be passed onto the ``write_method``.
    **csv_kwargs
        Any keyword arguments that will be passed onto csv.writer.
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, *args, **kwargs):
        self.initialise(kwargs)
        super().__init__(*args, **kwargs)


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class CsvDictSortedListChunks(CsvDictMixin, SortedListChunks):
    """Write files that are sorted on key where each file has at max
    chunksize rows.

    This implementation uses sortedcontainers rather than lists. This uses a
    ``csv.DictWriter`` so rows should be added as dictionaries. The ``header``
    is absolutely required.

    Parameters
    ----------
    chunk_dir : `str`
        The directory to write the chunk files to.
    key : `function`
        A key function that will return the values to sort on. This will be
        passed onto the sorting routine.
    chunksize : `int`, optional, default: `100000`
        The number of rows to store in memory before sorting and outputting to
        the chunk file.
    write_mode : `function`, optional, default: `wt`
        The method mode to use to write the chunk files. In most cases this
        should not be changed but it is there just in case the ``write_method``
        has any odd requirements.
    write_method : `function`, optional, default: `gzip.open`
        The method to use to write the chunk files.
    header : `list` of `str`, optional, default: `NoneType`
        A required header to write to the start of each chunk file.
    delete_on_error : `bool`, optional, default: `True`
        Delete all the chunk files if the chunking routine ends in an error.
    reverse : `bool`, optional, default: `False`
        Sort in reverse.
    chunk_prefix : `str`, optional, default: `NoneType`
        An optional prefix to add to the start of each chunk file name.
    chunk_suffix : `str`, optional, default: `.chunk.gz`
        An optional suffix to add to to the end of each chunk filename. Set to
        ``NoneType`` if you do not want a suffix.
    **write_kwargs
        Any keyword arguments that will be passed onto the ``write_method``.
    **csv_kwargs
        Any keyword arguments that will be passed onto csv.writer.
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, *args, **kwargs):
        self.initialise(kwargs)
        super().__init__(*args, **kwargs)


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class CsvDictExtendedChunks(CsvDictMixin, ExtendedChunks):
    """Write files that are sorted on key where each file can have variable
    number of rows possibly exceeding the chunksize value.

    This uses a ``csv.DictWriter`` so rows should be added as dictionaries. The
    ``header`` is absolutely required.

    Parameters
    ----------
    chunk_dir : `str`
        The directory to write the chunk files to.
    key : `function`
        A key function that will return the values to sort on. This will be
        passed onto the sorting routine.
    chunksize : `int`, optional, default: `100000`
        The number of rows to store in memory before sorting and outputting to
        the chunk file.
    write_mode : `function`, optional, default: `wt`
        The method mode to use to write the chunk files. In most cases this
        should not be changed but it is there just in case the ``write_method``
        has any odd requirements.
    write_method : `function`, optional, default: `gzip.open`
        The method to use to write the chunk files.
    header : `list` of `str`, optional, default: `NoneType`
        An required header to write to the start of each chunk file.
    delete_on_error : `bool`, optional, default: `True`
        Delete all the chunk files if the chunking routine ends in an error.
    reverse : `bool`, optional, default: `False`
        Sort in reverse.
    chunk_prefix : `str`, optional, default: `NoneType`
        An optional prefix to add to the start of each chunk file name.
    chunk_suffix : `str`, optional, default: `.chunk.gz`
        An optional suffix to add to to the end of each chunk filename. Set to
        ``NoneType`` if you do not want a suffix.
    **write_kwargs
        Any keyword arguments that will be passed onto the ``write_method``.
    **csv_kwargs
        Any keyword arguments that will be passed onto csv.writer.

    Notes
    -----
    This will keep adding chunks to a chunk file if the sort order is
    maintained between the last element of the previous chunk and the first
    element of the current chunk. This can reduce the number of chunk files
    that are needed and is useful if you suspect that the input may be sorted
    already.
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, *args, **kwargs):
        self.initialise(kwargs)
        super().__init__(*args, **kwargs)



"""For iterative merging of pre-sorted files.
"""
from merge_sort import common
from pyaddons import utils
from pyaddons.flat_files import header as pyhead
import gzip
import heapq
import os
import time
# import pprint as pp


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class IterativeHeapqMerge(object):
    """Performs an interative merge sort over many files.

    Parameters
    ----------
    files : `list` of `str`
        One or more file names. If header is ``True`` then the first row of
        each file must be the header.
    key : `function`, optional, default: `NoneType`
        The function to provide the sort keys to ``heapq.merge``.
    max_files : `int`, optional, default: `16`
        The maximum number of files that can be open (and merged) at any
        one time. The default (16) is the same as unix sort.
    tmpdir : `str`, optional, default: `NoneType`
        The directory for temp merge files to be placed if the merge can't be
        performed in a single pass. The default is the system temp location.
    read_method : `function`, optional, default: `gzip.open`
        The method to use to open the input files for merging.
    write_method : `function`, optional, default: `gzip.open`
        The method to use to write any output files if the merge can't be
        completed in a single pass.
    read_mode : `str`, optional, default: `rt`
        The mode to use to open the input files for merging.
    write_mode : `str`, optional, default: `wt`
        The mode to use to write and intermediate files if the merge cannot be
        completed in a single pass.
    header : `bool`, optional, default: `True`
        Do the input files have a header, if so, the header of the first file
        is stored in the ``header`` attribute.
    check_header : `bool`, optional, default: `True`
        If ``header`` is ``True``, do you want to make sure that all the
        headers are the same in each of the input files?
    delete : `bool`, optional, default: `False`
        Delete the source files as they are merged?
    reverse : `bool`, optional, default: `False`
        Perform the sort in reverse.
    comment : `str`, optional, default: `NoneType`
        A comment character that might appear before the header/data row, these
        are skipped. These are not checked beyond the first header/data row.
    read_kwargs : `dict`, optional, default: `NoneType`
        Any keyword arguments that will be passed to the ``read_method``.
    write_kwargs : `dict`, optional, default: `NoneType`
        Any keyword arguments that will be passed to the ``write_method``.
    """

    TEMP_PREFIX = "heapq_merge_"
    """A prefix that will be added to any intermediate merge files should the
    merge not be completed in a single pass (`str`).
    """

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, files, key=None, max_files=16, tmpdir=None,
                 read_method=gzip.open, write_method=gzip.open,
                 read_mode='rt', write_mode='wt', header=True,
                 check_header=True, delete=False, reverse=False,
                 comment=None, read_kwargs=None, write_kwargs=None):
        self._files = files
        self._key = key
        self._tmpdir = tmpdir
        self._read_method = read_method
        self._write_method = write_method
        self._read_mode = read_mode
        self._write_mode = write_mode
        self._read_kwargs = read_kwargs or dict()
        self._write_kwargs = write_kwargs or dict()
        self._has_header = header
        self._reverse = reverse
        self._delete = delete
        self._check_header = check_header
        self._comment = comment

        self._max_files = 0
        self.max_files = max_files

        # This will hold intermediate file names created during the merge
        # thee will be deleted at the end of the run.
        self._tempfiles = []
        self._remaining_files = []
        self._cur_time = time.time()
        self._temp_prefix = self.TEMP_PREFIX
        self.header = None
        self._final_merge = []

        # indicator to make sure we have opened the merger
        self._opened = False

        # Initialised to None so will be set up properly on a call to next()
        self._heapq = None

        # Will hold the iterations, number of files merged in each iteration
        # and time taken for each iteration.
        self._iterations = []

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __enter__(self):
        """Entry point for the context manager
        """
        self.open()
        return self

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __exit__(self, *args):
        """Exit point for the context manager

        Parameters
        ----------
        *args
            Any errors that are raised in the context manager, or  a tuple of
            3 `NoneType` if none were raised.
        """
        self.close()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __iter__(self):
        """Initialise the iterator.
        """
        return self

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __next__(self):
        """Grab the next row, rows are returned in sort order.

        Returns
        -------
        row : `list`
            All the entries are assumed to be bytes apart.
        """
        try:
            row = next(self._heapq)
        except TypeError:
            # The first call so we initialise, although we make sure that the
            # heapq is NoneType, if not then something else may have happened.
            if self._heapq is not None:
                raise

            if self._opened is False:
                raise IOError("did you forget to open the merge object?")

            # Otherwise setup the sort and return the first row
            self._merge()
            self._final_merge_start = time.time()
            try:
                row = next(self._heapq)
            except StopIteration:
                self._log_final_iteration()
                raise
        except StopIteration:
            self._log_final_iteration()
            raise

        return row

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def _log_final_iteration(self):
        """Log the timings for the final iteration.
        """
        try:
            iteration_no = self._iterations[-1][0]+1
        except IndexError:
            # no previous iterations it is doing it in one
            iteration_no = 1

        self._iterations.append(
            (
                iteration_no,
                len(self._final_merge),
                time.time() - self._final_merge_start,
                [i.name for i in self._final_merge]
            )
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def max_files(self):
        """Get the max open files files at one time (`int`)
        """
        return self._max_files

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @max_files.setter
    def max_files(self, max_files):
        """Set the max open files at one time but be > 1 (`int`)
        """
        max_files = int(max_files)
        if max_files < 2:
            raise ValueError(
                "must allow at least two files to be open at one time"
            )
        self._max_files = max_files

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def iterations(self):
        """Get the iterations used to merge the files (`list` of `tuple`).

        Each tuple is the iteration number, number of files, time for the
        iteration and files merged.
        """
        return self._iterations

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def open(self):
        """Open all the files that need merging and perform header checks.

        Notes
        -----
        Note that this will open and close all the files needed for merging to
        make sure they exist before the merge takes place. This is done as it
        can be time consuming, so we can fail due to header issues or
        FileNotFoundError at the start rather then during the process.
        """
        # If the files have headers
        if self._has_header is True:
            # Get the reference header row
            self.header = self.get_header(self._files[0])

            if self._check_header is True:
                # Make sure all the files can be opened and have the same
                # header
                for i in self._files[1:]:
                    comp_header = self.get_header(i)
                    if self.header != self.get_header(i):
                        raise ValueError(
                            f"headers are different: {self._files[0]} vs. {i}"
                            f"[{self.header}] vs. [{comp_header}]"
                        )
            else:
                # If we are not checking the header just make sure we can open
                for i in self._files:
                    open(i).close()
        else:
            # We have no header so just make sure we can open
            for i in self._files:
                open(i).close()

        # Indicate that open has been called
        self._opened = True

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def close(self):
        """Close all the files, and ensure any files that need deleting are
        removed.
        """
        self.close_file_batch(self._final_merge)

        for i in self._tempfiles:
            try:
                os.unlink(i)
            except Exception:
                pass

        if self._delete is True:
            for i in self._remaining_files:
                try:
                    os.unlink(i)
                except Exception:
                    pass

        self._opened = False

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_header(self, filename):
        """Grab a header row from a input file. This will open read the first
        row and close the file.

        Parameters
        ----------
        filename : `str`
            The path to the file.

        Returns
        -------
        header : `str`
            The first row from the file.
        """
        with self.open_file(filename) as infile:
            return next(infile)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def open_file(self, filename):
        """Open a file for reading. This will also skip any fixed numbers of
        leading lines and comment characters.

        Parameters
        ----------
        filename : `str`
            The path to the file.

        Returns
        -------
        file_obj : `File`
            A file object for reading.
        """
        file_obj = self._read_method(
            filename, self._read_mode, **self._read_kwargs
        )

        # If we have leading comment characters or fixed numbers of lines to
        # skip then do so before returning the file object.
        if self._comment is not None:
            pyhead.move_to_header(
                file_obj, comment=self._comment
            )
        return file_obj

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def open_file_batch(self, files):
        """Open a batch of files for reading and merging.

        Parameters
        ----------
        files : `list` of `str`
            A list of files to open.

        Returns
        -------
        file_objs : `list` or `File`
            A list of open file objects.
        """
        # The final merge
        file_objs = [
            self.open_file(i) for i in files
        ]
        return file_objs

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def close_file_batch(self, files):
        """Open a batch of files for reading and merging.

        Parameters
        ----------
        files : `list` of `File`
            A list of files objects to close.
        """
        for i in files:
            i.close()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def read_file_batch_header(self, files):
        """Read a single row from a batch of files.

        Parameters
        ----------
        files : `list` of `File`
            A list of files objects to read from.
        """
        for i in files:
            next(i)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def open_merge_file(self):
        """Open a merge file file for writing.

        Returns
        -------
        merge_file : `File`
            A file object to write an intermediate merge file to.
        """
        # This will hold the sorted output for the current loop
        tmp_merge = utils.get_tmp_file(
            dir=self._tmpdir, prefix=self._temp_prefix
        )
        self._tempfiles.append(tmp_merge)

        return self._write_method(
            tmp_merge, self._write_mode, **self._write_kwargs
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def write_header(self, file):
        """Write a header row to a file.

        Returns
        -------
        file : `File`
            A file object to write the header to.
        """
        file.write(self.header)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def _merge(self):
        """Start the merge process until we have a small enough remaining
        files to merge in a single pass.
        """
        # This will hold all the files that need merging over the various
        # iterations
        self._remaining_files = list(self._files)

        # Holds the file names on intermediate (so files that are created
        # because we can't merge all the files in one go). These will be
        # deleted irrespective of the overall delete parameter
        self._tempfiles = []

        iterations = 0

        # We merge until we get below are our file limit and then hopefully
        # we will have a one final merge
        while len(self._remaining_files) > self._max_files:
            start_at = time.time()
            to_merge = []
            for i in range(self._max_files):
                try:
                    # Fill up from the start of the file list
                    to_merge.append(self._remaining_files.pop(0))
                except IndexError:
                    # No more files remaining
                    break

            merge_file = self.open_merge_file()

            try:
                merge_fobjs = self.open_file_batch(to_merge)

                # Now we merge into the temp file
                if self._has_header is True:
                    self.read_file_batch_header(merge_fobjs)
                    self.write_header(merge_file)

                # Now merge the rest
                for row in heapq.merge(*merge_fobjs, key=self._key,
                                       reverse=self._reverse):
                    merge_file.write(row)
            except Exception:
                merge_file.close()
                self.close_file_batch(merge_fobjs)
                raise

            merge_file_name = merge_file.name
            merge_file.close()
            self.close_file_batch(merge_fobjs)

            # Also the tmp_file we just created is now remaining
            self._remaining_files.append(merge_file_name)

            # Delete the batch
            if self._delete is True:
                for i in to_merge:
                    os.unlink(i)

            iterations += 1
            end_at = time.time()
            self._iterations.append(
                (iterations, len(merge_fobjs), end_at - start_at,
                 [i.name for i in merge_fobjs])
            )

        # The final merge
        self._final_merge = self.open_file_batch(self._remaining_files)

        # Now we merge into the temp file
        if self._has_header is True:
            self.read_file_batch_header(self._final_merge)

        self._heapq = heapq.merge(
            *self._final_merge,
            key=self._key,
            reverse=self._reverse
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class CsvIterativeHeapqMerge(IterativeHeapqMerge):
    """Performs an interative merge sort over many files.

    Parameters
    ----------
    files : `list` of `str`
        One or more file names. If header is ``True`` then the first row of
        each file must be the header.
    key : `function`, optional, default: `NoneType`
        The function to provide the sort keys to ``heapq.merge``.
    max_files : `int`, optional, default: `16`
        The maximum number of files that can be open (and merged) at any
        one time. The default (16) is the same as unix sort.
    tmpdir : `str`, optional, default: `NoneType`
        The directory for temp merge files to be placed if the merge can't be
        performed in a single pass. The default is the system temp location.
    read_method : `function`, optional, default: `gzip.open`
        The method to use to open the input files for merging.
    write_method : `function`, optional, default: `gzip.open`
        The method to use to write any output files if the merge can't be
        completed in a single pass.
    read_mode : `str`, optional, default: `rt`
        The mode to use to open the input files for merging.
    write_mode : `str`, optional, default: `wt`
        The mode to use to write and intermediate files if the merge cannot be
        completed in a single pass.
    header : `bool`, optional, default: `True`
        Do the input files have a header, if so, the header of the first file
        is stored in the ``header`` attribute.
    check_header : `bool`, optional, default: `True`
        If ``header`` is ``True``, do you want to make sure that all the
        headers are the same in each of the input files?
    delete : `bool`, optional, default: `False`
        Delete the source files as they are merged?
    reverse : `bool`, optional, default: `False`
        Perform the sort in reverse.
    comment : `str`, optional, default: `NoneType`
        A comment character that might appear before the header/data row, these
        are skipped. These are not checked beyond the first header/data row.
    read_kwargs : `dict`, optional, default: `NoneType`
        Any keyword arguments that will be passed to the ``read_method``.
    write_kwargs : `dict`, optional, default: `NoneType`
        Any keyword arguments that will be passed to the ``write_method``.
    csv_kwargs : `dict`, optional, default: `NoneType`
        Any keyword arguments that will be passed to the
        ``csv.reader``/``csv.writer``.
    """

    TEMP_PREFIX = "csv_heapq_merge_"
    """A prefix that will be added to any intermediate merge files should the
    merge not be completed in a single pass (`str`).
    """

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, *args, csv_kwargs=None, **kwargs):
        super().__init__(*args, **kwargs)
        csv_kwargs = csv_kwargs or dict()
        self._csv_kwargs = csv_kwargs

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def open_file(self, filename):
        """Open a file for reading.

        Parameters
        ----------
        filename : `str`
            The path to the file.

        Returns
        -------
        csv_wrapper : `merge_sort.common.CsvWrapper`
            A file object for reading.
        """
        file_obj = super().open_file(filename)
        csv_reader = common.CsvWrapper(file_obj, mode='r', **self._csv_kwargs)
        return csv_reader.open()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def open_merge_file(self):
        """Open a merge file file for writing.

        Returns
        -------
        csv_wrapper : `merge_sort.common.CsvWrapper`
            A file object for writing.
        """
        file_obj = super().open_merge_file()
        csv_reader = common.CsvWrapper(file_obj, mode='w', **self._csv_kwargs)
        return csv_reader.open()


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class CsvDictIterativeHeapqMerge(IterativeHeapqMerge):
    """Performs an interative merge sort over many files.

    Parameters
    ----------
    files : `list` of `str`
        One or more file names. If header is ``True`` then the first row of
        each file must be the header.
    key : `function`, optional, default: `NoneType`
        The function to provide the sort keys to ``heapq.merge``.
    max_files : `int`, optional, default: `16`
        The maximum number of files that can be open (and merged) at any
        one time. The default (16) is the same as unix sort.
    tmpdir : `str`, optional, default: `NoneType`
        The directory for temp merge files to be placed if the merge can't be
        performed in a single pass. The default is the system temp location.
    read_method : `function`, optional, default: `gzip.open`
        The method to use to open the input files for merging.
    write_method : `function`, optional, default: `gzip.open`
        The method to use to write any output files if the merge can't be
        completed in a single pass.
    read_mode : `str`, optional, default: `rt`
        The mode to use to open the input files for merging.
    write_mode : `str`, optional, default: `wt`
        The mode to use to write and intermediate files if the merge cannot be
        completed in a single pass.
    header : `bool`, optional, default: `True`
        Do the input files have a header, if so, the header of the first file
        is stored in the ``header`` attribute.
    check_header : `bool`, optional, default: `True`
        If ``header`` is ``True``, do you want to make sure that all the
        headers are the same in each of the input files?
    delete : `bool`, optional, default: `False`
        Delete the source files as they are merged?
    reverse : `bool`, optional, default: `False`
        Perform the sort in reverse.
    comment : `str`, optional, default: `NoneType`
        A comment character that might appear before the header/data row, these
        are skipped. These are not checked beyond the first header/data row.
    read_kwargs : `dict`, optional, default: `NoneType`
        Any keyword arguments that will be passed to the ``read_method``.
    write_kwargs : `dict`, optional, default: `NoneType`
        Any keyword arguments that will be passed to the ``write_method``.
    csv_kwargs : `dict`, optional, default: `NoneType`
        Any keyword arguments that will be passed to the
        ``csv.DictReader``/``csv.DictWriter``.
    """

    TEMP_PREFIX = "csv_dict_heapq_merge_"
    """A prefix that will be added to any intermediate merge files should the
    merge not be completed in a single pass (`str`).
    """

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, *args, csv_kwargs=None, **kwargs):
        super().__init__(*args, **kwargs)
        csv_kwargs = csv_kwargs or dict()
        self._csv_read_kwargs = csv_kwargs
        self._csv_write_kwargs = dict(csv_kwargs)

        # Remove any DictWriter only keyword arguments
        for i in ['restval', 'extrasaction']:
            try:
                self._csv_read_kwargs.pop(i)
            except KeyError:
                pass

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_header(self, filename):
        """Grab a header row from a input file. This will open read the first
        row and close the file.

        Parameters
        ----------
        filename : `str`
            The path to the file.

        Returns
        -------
        header : `str`
            The first row from the file.
        """
        try:
            infile = self.open_file(filename)
            header = infile.csv.fieldnames
        finally:
            infile.close()
        return header

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def read_file_batch_header(self, files):
        """Read a single row from a batch of files.

        Parameters
        ----------
        files : `list` of `File`
            A list of files objects to read from.
        """
        pass

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def open_file(self, filename):
        """Open a file for reading.

        Parameters
        ----------
        filename : `str`
            The path to the file.

        Returns
        -------
        csv_wrapper : `merge_sort.common.CsvDictWrapper`
            A file object for reading.
        """
        file_obj = super().open_file(filename)
        csv_reader = common.CsvDictWrapper(
            file_obj, mode='r', **self._csv_read_kwargs
        )
        return csv_reader.open()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def open_merge_file(self):
        """Open a merge file file for writing.

        Returns
        -------
        csv_wrapper : `merge_sort.common.CsvDictWrapper`
            A file object for writing.
        """
        file_obj = super().open_merge_file()
        csv_reader = common.CsvDictWrapper(
            file_obj, mode='w', fieldnames=self.header,
            **self._csv_write_kwargs
        )
        return csv_reader.open()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def write_header(self, file):
        """Writer a header row to the file.
        """
        file.csv.writeheader()
